#include "mainwindow.h"
#include "chart.h"
#include "chartview.h"
#include "smartfield_fileloader.h"
#include "ui_mainwindow.h"

#include <QAction>
#include <QFileDialog>
#include <QMessageBox>
#include <complex.h>

MainWindow::MainWindow(QWidget* parent)
    : QMainWindow(parent)
    , ui(new Ui::MainWindow)
{
    ui->setupUi(this);
    ui->treeWidget->setHeaderHidden(true);

    connect(ui->treeWidget, SIGNAL(itemExpanded(QTreeWidgetItem*)), this, SLOT(treeWidgetItemExpanded(QTreeWidgetItem*)));
    connect(ui->actionOpen, SIGNAL(triggered()), this, SLOT(openFiles()));
    connect(ui->actionCloseCharts, SIGNAL(triggered()), this, SLOT(closeChartWindows()));
    connect(ui->pushButtonRememberUIDCalib, SIGNAL(clicked()), this, SLOT(rememberUIDCalib()));
    connect(signalMapperRawData, SIGNAL(mapped(int)), this, SLOT(rawDataChart(int)));
    connect(signalMapperUTIT, SIGNAL(mapped(int)), this, SLOT(filteredDataChart(int)));

    QList<QString>* colNames = new QList<QString>();
    colNames->append("Číslo měření");
    colNames->append("Vrms [V]");
    colNames->append("Irms [mA]");
    colNames->append("φ [°]");
    colNames->append("Rp [Ω]");
    colNames->append("Cp [pF]");
    colNames->append("Rntc [Ω]");
    colNames->append("Temperature [°C]");

    ui->tableResults_FIR->setRowCount(1);
    ui->tableResults_FIR->setColumnCount(colNames->length() + 2);
    for (int i = 0; i < colNames->length(); i++) {
        QVariant oVariant(colNames->at(i));
        // allocate the widget item
        QTableWidgetItem* poItem = new QTableWidgetItem();
        poItem->setTextAlignment(Qt::AlignCenter);
        poItem->setData(Qt::DisplayRole, oVariant);
        ui->tableResults_FIR->setItem(0, i, poItem);
    }
    ui->tableResults_FIR->setRowHeight(0, 1.5 * ui->tableResults_FIR->rowHeight(0));
    ui->tableResults_FIR->setColumnWidth(0, 0.8 * ui->tableResults_FIR->columnWidth(0));

    ui->tableResults_DFT->setRowCount(1);
    ui->tableResults_DFT->setColumnCount(colNames->length() + 2);
    for (int i = 0; i < colNames->length(); i++) {
        QVariant oVariant(colNames->at(i));
        // allocate the widget item
        QTableWidgetItem* poItem = new QTableWidgetItem();
        poItem->setTextAlignment(Qt::AlignCenter);
        poItem->setData(Qt::DisplayRole, oVariant);
        ui->tableResults_DFT->setItem(0, i, poItem);
    }
    ui->tableResults_DFT->setRowHeight(0, 1.5 * ui->tableResults_DFT->rowHeight(0));
    ui->tableResults_DFT->setColumnWidth(0, 0.8 * ui->tableResults_DFT->columnWidth(0));

    //btn_edit->setText("u(t),i(t)");
}

MainWindow::~MainWindow()
{
    delete ui;
}

void MainWindow::fillTableOfResults()
{
    /* Clear previous content */
    for (int i = ui->tableResults_FIR->rowCount() - 1; i > 0; i--) {
        ui->tableResults_FIR->removeRow(i);
    }

    ui->tableResults_FIR->setRowCount(ui->tableResults_FIR->rowCount() + sfLoader.getNumberOfMeasurements() + 2);

    float average_vrms = 0;
    float average_irms = 0;
    float average_angle = 0;
    float _Complex average_zc = 0;

    for (int i = 0; i < sfLoader.getNumberOfMeasurements(); i++) {
        QTableWidgetItem* poItem = new QTableWidgetItem();
        poItem->setData(Qt::DisplayRole, QVariant(i + 1));
        poItem->setTextAlignment(Qt::AlignCenter);
        ui->tableResults_FIR->setItem(i + 1, 0, poItem);

        SmartField_Result_t result = sfSIM.results.at(i);

        average_vrms += result.VRMS;
        average_irms += result.IRMS;
        average_angle += result.angle;
        average_zc += result.ZC;

        poItem = new QTableWidgetItem();
        poItem->setData(Qt::DisplayRole, QVariant(result.VRMS));
        poItem->setTextAlignment(Qt::AlignCenter);
        ui->tableResults_FIR->setItem(i + 1, 1, poItem);

        poItem = new QTableWidgetItem();
        poItem->setData(Qt::DisplayRole, QVariant(result.IRMS * 1e3));
        poItem->setTextAlignment(Qt::AlignCenter);
        ui->tableResults_FIR->setItem(i + 1, 2, poItem);

        poItem = new QTableWidgetItem();
        poItem->setData(Qt::DisplayRole, QVariant(result.angle));
        poItem->setTextAlignment(Qt::AlignCenter);
        ui->tableResults_FIR->setItem(i + 1, 3, poItem);

        poItem = new QTableWidgetItem();
        poItem->setData(Qt::DisplayRole, QVariant(1 / creal(1 / result.ZC)));
        poItem->setTextAlignment(Qt::AlignCenter);
        ui->tableResults_FIR->setItem(i + 1, 4, poItem);

        poItem = new QTableWidgetItem();
        poItem->setData(Qt::DisplayRole, QVariant(1e12 * cimag(1 / result.ZC) / (2 * M_PI * 10e6)));
        poItem->setTextAlignment(Qt::AlignCenter);
        ui->tableResults_FIR->setItem(i + 1, 5, poItem);

        QWidget* pWidget;
        QPushButton* button;
        QHBoxLayout* pLayout;

        pWidget = new QWidget();
        button = new QPushButton();
        button->setText("Raw data");
        pLayout = new QHBoxLayout(pWidget);
        pLayout->addWidget(button);
        pLayout->setAlignment(Qt::AlignCenter);
        pLayout->setContentsMargins(0, 0, 0, 0);
        pWidget->setLayout(pLayout);
        ui->tableResults_FIR->setCellWidget(i + 1, 8, pWidget);

        connect(button, SIGNAL(clicked()), signalMapperRawData, SLOT(map()));
        signalMapperRawData->setMapping(button, i);

        pWidget = new QWidget();
        button = new QPushButton();
        button->setText("u(t), i(t)");
        pLayout = new QHBoxLayout(pWidget);
        pLayout->addWidget(button);
        pLayout->setAlignment(Qt::AlignCenter);
        pLayout->setContentsMargins(0, 0, 0, 0);
        pWidget->setLayout(pLayout);
        ui->tableResults_FIR->setCellWidget(i + 1, 9, pWidget);

        connect(button, SIGNAL(clicked()), signalMapperUTIT, SLOT(map()));
        signalMapperUTIT->setMapping(button, i);
    }

    //Fill in averages
    uint32_t i = sfLoader.getNumberOfMeasurements();
    average_vrms /= i;
    average_irms /= i;
    average_angle /= i;
    average_zc /= i;

    i++;
    QTableWidgetItem* poItem = new QTableWidgetItem();
    poItem->setData(Qt::DisplayRole, QVariant("Průměr"));
    poItem->setTextAlignment(Qt::AlignCenter);
    ui->tableResults_FIR->setItem(ui->tableResults_FIR->rowCount() - 1, 0, poItem);

    poItem = new QTableWidgetItem();
    poItem->setData(Qt::DisplayRole, QVariant(average_vrms));
    poItem->setTextAlignment(Qt::AlignCenter);
    ui->tableResults_FIR->setItem(i + 1, 1, poItem);

    poItem = new QTableWidgetItem();
    poItem->setData(Qt::DisplayRole, QVariant(average_irms * 1e3));
    poItem->setTextAlignment(Qt::AlignCenter);
    ui->tableResults_FIR->setItem(i + 1, 2, poItem);

    poItem = new QTableWidgetItem();
    poItem->setData(Qt::DisplayRole, QVariant(average_angle));
    poItem->setTextAlignment(Qt::AlignCenter);
    ui->tableResults_FIR->setItem(i + 1, 3, poItem);

    poItem = new QTableWidgetItem();
    poItem->setData(Qt::DisplayRole, QVariant(1 / creal(1 / average_zc)));
    poItem->setTextAlignment(Qt::AlignCenter);
    ui->tableResults_FIR->setItem(i + 1, 4, poItem);

    poItem = new QTableWidgetItem();
    poItem->setData(Qt::DisplayRole, QVariant(1e12 * cimag(1 / average_zc) / (2 * M_PI * 10e6)));
    poItem->setTextAlignment(Qt::AlignCenter);
    ui->tableResults_FIR->setItem(i + 1, 5, poItem);
}

void MainWindow::fillTableOfResults_DFT()
{
    /* Clear previous content */
    for (int i = ui->tableResults_DFT->rowCount() - 1; i > 0; i--) {
        ui->tableResults_DFT->removeRow(i);
    }

    ui->tableResults_DFT->setRowCount(ui->tableResults_DFT->rowCount() + sfLoader.getNumberOfMeasurements() + 2);

    float average_vrms = 0;
    float average_irms = 0;
    float average_angle = 0;
    float _Complex average_zc = 0;

    for (int i = 0; i < sfLoader.getNumberOfMeasurements(); i++) {
        QTableWidgetItem* poItem = new QTableWidgetItem();
        poItem->setData(Qt::DisplayRole, QVariant(i + 1));
        poItem->setTextAlignment(Qt::AlignCenter);
        ui->tableResults_DFT->setItem(i + 1, 0, poItem);

        SmartField_Result_t result = sfSIM.results_DFT.at(i);

        average_vrms += result.VRMS;
        average_irms += result.IRMS;
        average_angle += result.angle;
        average_zc += result.ZC;

        poItem = new QTableWidgetItem();
        poItem->setData(Qt::DisplayRole, QVariant(result.VRMS));
        poItem->setTextAlignment(Qt::AlignCenter);
        ui->tableResults_DFT->setItem(i + 1, 1, poItem);

        poItem = new QTableWidgetItem();
        poItem->setData(Qt::DisplayRole, QVariant(result.IRMS * 1e3));
        poItem->setTextAlignment(Qt::AlignCenter);
        ui->tableResults_DFT->setItem(i + 1, 2, poItem);

        poItem = new QTableWidgetItem();
        poItem->setData(Qt::DisplayRole, QVariant(result.angle));
        poItem->setTextAlignment(Qt::AlignCenter);
        ui->tableResults_DFT->setItem(i + 1, 3, poItem);

        poItem = new QTableWidgetItem();
        poItem->setData(Qt::DisplayRole, QVariant(1 / creal(1 / result.ZC)));
        poItem->setTextAlignment(Qt::AlignCenter);
        ui->tableResults_DFT->setItem(i + 1, 4, poItem);

        poItem = new QTableWidgetItem();
        poItem->setData(Qt::DisplayRole, QVariant(1e12 * cimag(1 / result.ZC) / (2 * M_PI * 10e6)));
        poItem->setTextAlignment(Qt::AlignCenter);
        ui->tableResults_DFT->setItem(i + 1, 5, poItem);

        QWidget* pWidget;
        QPushButton* button;
        QHBoxLayout* pLayout;

        pWidget = new QWidget();
        button = new QPushButton();
        button->setText("Raw data");
        pLayout = new QHBoxLayout(pWidget);
        pLayout->addWidget(button);
        pLayout->setAlignment(Qt::AlignCenter);
        pLayout->setContentsMargins(0, 0, 0, 0);
        pWidget->setLayout(pLayout);
        ui->tableResults_DFT->setCellWidget(i + 1, 8, pWidget);

        connect(button, SIGNAL(clicked()), signalMapperRawData, SLOT(map()));
        signalMapperRawData->setMapping(button, i);
    }

    //Fill in averages
    uint32_t i = sfLoader.getNumberOfMeasurements();
    average_vrms /= i;
    average_irms /= i;
    average_angle /= i;
    average_zc /= i;

    i++;
    QTableWidgetItem* poItem = new QTableWidgetItem();
    poItem->setData(Qt::DisplayRole, QVariant("Průměr"));
    poItem->setTextAlignment(Qt::AlignCenter);
    ui->tableResults_DFT->setItem(ui->tableResults_FIR->rowCount() - 1, 0, poItem);

    poItem = new QTableWidgetItem();
    poItem->setData(Qt::DisplayRole, QVariant(average_vrms));
    poItem->setTextAlignment(Qt::AlignCenter);
    ui->tableResults_DFT->setItem(i + 1, 1, poItem);

    poItem = new QTableWidgetItem();
    poItem->setData(Qt::DisplayRole, QVariant(average_irms * 1e3));
    poItem->setTextAlignment(Qt::AlignCenter);
    ui->tableResults_DFT->setItem(i + 1, 2, poItem);

    poItem = new QTableWidgetItem();
    poItem->setData(Qt::DisplayRole, QVariant(average_angle));
    poItem->setTextAlignment(Qt::AlignCenter);
    ui->tableResults_DFT->setItem(i + 1, 3, poItem);

    poItem = new QTableWidgetItem();
    poItem->setData(Qt::DisplayRole, QVariant(1 / creal(1 / average_zc)));
    poItem->setTextAlignment(Qt::AlignCenter);
    ui->tableResults_DFT->setItem(i + 1, 4, poItem);

    poItem = new QTableWidgetItem();
    poItem->setData(Qt::DisplayRole, QVariant(1e12 * cimag(1 / average_zc) / (2 * M_PI * 10e6)));
    poItem->setTextAlignment(Qt::AlignCenter);
    ui->tableResults_DFT->setItem(i + 1, 5, poItem);
}

void MainWindow::rawDataChart(int measurement)
{
    if (measurement >= sfLoader.voltage.length()) {
        return;
    }

    QLineSeries* vSeries = new QLineSeries();
    QLineSeries* iSeries = new QLineSeries();

    for (int i = 0; i < sfLoader.voltage.at(measurement).length(); i++) {
        *vSeries << QPointF(i, sfLoader.voltage.at(measurement).at(i));
        *iSeries << QPointF(i, sfLoader.current.at(measurement).at(i));
    }

    vSeries->setColor(Qt::blue);
    vSeries->setName("Napětí [-]");

    iSeries->setColor(Qt::red);
    iSeries->setName("Proud [-]");

    QFileInfo inFile = QFileInfo(QFile(ui->filesList->currentItem()->text()));
    QString chartTitle;
    chartTitle += "Surová data ze souboru: <b>" + inFile.baseName() + "</b>." + inFile.completeSuffix() + "<br>";
    chartTitle += "UID jednotky: " + sfLoader.getUID() + ", Měření č. <b>" + QString::number(measurement + 1) + "</b>";

    QValueAxis* axisX = new QValueAxis;
    axisX->setRange(0, vSeries->points().length() - 1);
    axisX->setTickCount(9);
    axisX->setLabelFormat("%d");

    QValueAxis* axisY = new QValueAxis;
    axisY->setRange(0, 255);
    axisY->setTickCount(9);
    axisY->setLabelFormat("%d");

    Chart* chart = new Chart();
    chart->addSeries(vSeries);
    chart->addSeries(iSeries);
    chart->setTitle(chartTitle);
    chart->setAnimationOptions(QChart::NoAnimation);
    //chart->legend()->hide();
    chart->legend()->show();
    chart->addAxis(axisX, Qt::AlignBottom);
    chart->addAxis(axisY, Qt::AlignLeft);
    vSeries->attachAxis(axisX);
    vSeries->attachAxis(axisY);
    iSeries->attachAxis(axisX);
    iSeries->attachAxis(axisY);

    ChartView* chartView = new ChartView(chart);
    chartView->setRenderHint(QPainter::Antialiasing);

    //QWidget* window = new QWidget;
    QMainWindow* window = new QMainWindow;
    window->setWindowIcon(QIcon(":/icon.png"));
    window->setCentralWidget(chartView);
    window->resize(800, 600);
    window->grabGesture(Qt::PanGesture);
    window->grabGesture(Qt::PinchGesture);
    window->show();

    chartWindows->append(window);
}

void MainWindow::filteredDataChart(int measurement)
{
    if (measurement >= sfLoader.voltage.length()) {
        return;
    }

    QLineSeries* vSeries = new QLineSeries();
    QLineSeries* iSeries = new QLineSeries();

    for (int i = 0; i < sfSIM.voltage.at(measurement).length(); i++) {
        *vSeries << QPointF(i, sfSIM.voltage.at(measurement).at(i));
        *iSeries << QPointF(i, sfSIM.current.at(measurement).at(i) * 1e3);
    }

    vSeries->setColor(Qt::blue);
    vSeries->setName("Napětí [V]");

    iSeries->setColor(Qt::red);
    iSeries->setName("Proud [mA]");

    QFileInfo inFile = QFileInfo(QFile(ui->filesList->currentItem()->text()));
    QString chartTitle;
    chartTitle += "Filtrovaná data ze souboru: <b>" + inFile.baseName() + "</b>." + inFile.completeSuffix() + "<br>";
    chartTitle += "UID jednotky: " + sfLoader.getUID() + ", Měření č. <b>" + QString::number(measurement + 1) + "</b>";

    QValueAxis* axisX = new QValueAxis;
    axisX->setRange(0, vSeries->points().length() - 1);
    axisX->setTickCount(9);
    axisX->setLabelFormat("%d");

    QValueAxis* axisY1 = new QValueAxis;
    float voltageMin = *std::min_element(sfSIM.voltage.at(measurement).begin(), sfSIM.voltage.at(measurement).end());
    float voltageMax = *std::max_element(sfSIM.voltage.at(measurement).begin(), sfSIM.voltage.at(measurement).end());
    float scale = (voltageMax >= -voltageMin) ? voltageMax : -voltageMin;
    axisY1->setRange(-scale, scale);
    axisY1->setTickCount(9);
    axisY1->setLabelFormat("%0.2f");
    axisY1->setLinePenColor(Qt::blue);

    QValueAxis* axisY2 = new QValueAxis;
    float curentMin = *std::min_element(sfSIM.current.at(measurement).begin(), sfSIM.current.at(measurement).end());
    float curentMax = *std::max_element(sfSIM.current.at(measurement).begin(), sfSIM.current.at(measurement).end());
    scale = (curentMax >= -curentMin) ? curentMax : -curentMin;
    scale = scale * 1e3;
    axisY2->setRange(-scale, scale);
    axisY2->setTickCount(9);
    axisY2->setLabelFormat("%0.2f");
    axisY2->setLinePenColor(Qt::red);

    Chart* chart = new Chart();
    chart->addSeries(vSeries);
    chart->addSeries(iSeries);
    chart->setTitle(chartTitle);
    chart->setAnimationOptions(QChart::NoAnimation);
    //chart->legend()->hide();
    chart->legend()->show();
    //    chart->createDefaultAxes();
    //    chart->axes(Qt::Horizontal).clear();
    chart->addAxis(axisX, Qt::AlignBottom);
    chart->addAxis(axisY1, Qt::AlignLeft);
    chart->addAxis(axisY2, Qt::AlignRight);
    vSeries->attachAxis(axisX);
    vSeries->attachAxis(axisY1);
    iSeries->attachAxis(axisX);
    iSeries->attachAxis(axisY2);

    ChartView* chartView = new ChartView(chart);
    chartView->setRenderHint(QPainter::Antialiasing);

    //QWidget* window = new QWidget;
    QMainWindow* window = new QMainWindow;
    window->setWindowIcon(QIcon(":/icon.png"));
    window->setCentralWidget(chartView);
    window->resize(800, 600);
    window->grabGesture(Qt::PanGesture);
    window->grabGesture(Qt::PinchGesture);
    window->show();

    chartWindows->append(window);
}

void MainWindow::openFiles()
{
    QFileDialog dialog(this);
    dialog.setFileMode(QFileDialog::ExistingFiles);
    dialog.setNameFilter(tr("SmartField bianry file (*.bin);;All Files (*)"));
    dialog.setViewMode(QFileDialog::Detail);

    QStringList fileNames;
    if (dialog.exec()) {
        fileNames = dialog.selectedFiles();
    }

    ui->filesList->addItems(fileNames);
}

void MainWindow::on_filesList_currentItemChanged(QListWidgetItem* current, QListWidgetItem* previous)
{
    Q_UNUSED(previous)
    //    if (sfLoader != nullptr) {
    //        delete (sfLoader);
    //        sfLoader = nullptr;
    //    }
    //sfLoader->~SmartField_FileLoader();
    //sfLoader =  SmartField_FileLoader();
    sfLoader.reset();
    sfSIM.reset();
    sfLoader.loadFile(new QFile(current->text()));

    //Load default calibration values (Covers N/A identifier and unknown UID)
    ui->doubleSpinBox_phiMul->setValue(1);
    ui->doubleSpinBox_phiOffset->setValue(0);
    ui->doubleSpinBox_zMul->setValue(1);
    ui->doubleSpinBox_zOffset->setValue(0);

    if (sfLoader.getUID() != "N/A") {
        ui->pushButtonRememberUIDCalib->setEnabled(true);
        recallUIDCalib();
    }

    QLabel* label;
    QString labelText;
    label = ui->labelUID;
    label->setText(sfLoader.getUID());
    label->adjustSize();

    label = ui->labelHeaderVersion;
    label->setText(QString::number(sfLoader.getHeaderVersion()));
    label->adjustSize();

    label = ui->labelNumberOfMeasurements;
    label->setText(QString::number(sfLoader.getNumberOfMeasurements()));
    label->adjustSize();

    label = ui->labelNumberOfPoints;
    label->setText(QString::number(sfLoader.getNumberOfPoints()));
    label->adjustSize();

    label = ui->labelNTCbeta;
    label->setText(QString::number(sfLoader.getNTCbeta()));
    label->adjustSize();

    label = ui->labelZsns;
    label->setText(QString::number(sfLoader.getZsns(), 'f', 3) + " Ω");
    label->adjustSize();

    label = ui->labelNormalisedFrequency;
    label->setText(QString::number(2 * sfLoader.getNormalisedFreqeuncy(), 'f', 5) + " π");
    label->adjustSize();

    sfSIM.evaluateAll();
    fillTableOfResults();
    fillTableOfResults_DFT();
}

void MainWindow::rememberUIDCalib()
{
    QFile calFile("calib.json");
    QFileInfo fileInfo = QFileInfo(calFile);

    /* Prepare updated calibration data */
    QJsonObject calJson;
    calJson.insert("phiMul", ui->doubleSpinBox_phiMul->value());
    calJson.insert("phiOffset", ui->doubleSpinBox_phiOffset->value());
    calJson.insert("zMul", ui->doubleSpinBox_zMul->value());
    calJson.insert("zOffset", ui->doubleSpinBox_zOffset->value());

    /* Instance wrapper JSON */
    QJsonObject masterJson;

    if (fileInfo.exists()) {
        /* Populate wrapper JSON if file already exists */
        calFile.open(QIODevice::ReadWrite);
        QJsonDocument json = QJsonDocument::fromJson(calFile.readAll());
        calFile.seek(0);
        masterJson = json.object();
    } else {
        calFile.open(QIODevice::WriteOnly);
    }

    /* Insert/Overwrite calibration based on UID */
    masterJson.insert(sfLoader.getUID(), calJson);
    QJsonDocument json(masterJson);
    calFile.write(json.toJson());
    calFile.close();
}

void MainWindow::recallUIDCalib()
{
    QFile calFile("calib.json");
    QFileInfo fileInfo = QFileInfo(calFile);
    if (fileInfo.exists()) {
        calFile.open(QIODevice::ReadOnly);
        QJsonDocument json = QJsonDocument::fromJson(calFile.readAll());
        if (json.object().contains(sfLoader.getUID())) {
            //IF UID was found in JSON, load stored values
            QJsonObject oJson = json.object().value(sfLoader.getUID()).toObject();
            ui->doubleSpinBox_phiMul->setValue(oJson.value("phiMul").toDouble(1));
            ui->doubleSpinBox_phiOffset->setValue(oJson.value("phiOffset").toDouble(0));
            ui->doubleSpinBox_zMul->setValue(oJson.value("zMul").toDouble(1));
            ui->doubleSpinBox_zOffset->setValue(oJson.value("zOffset").toDouble(0));
        }
    }
}

void MainWindow::closeEvent(QCloseEvent* event)
{
    closeChartWindows();
    event->accept();
}

void MainWindow::closeChartWindows()
{
    /* Close all chart windows */
    for (QMainWindow* w : *chartWindows) {
        w->close();
        w->deleteLater();
    }
}

void MainWindow::treeWidgetItemExpanded(QTreeWidgetItem* item)
{
    Q_UNUSED(item);
    ui->treeWidget->resizeColumnToContents(0);
}
