#ifndef SMARTFIELD_FILELOADER_H
#define SMARTFIELD_FILELOADER_H

#include <QtCharts>

class SmartField_FileLoader {
public:
    static SmartField_FileLoader& getInstance()
    {
        static SmartField_FileLoader instance; // Guaranteed to be destroyed, instantiated on first use.
        return instance;
    }
    /* Prevent copies of singleton */
    SmartField_FileLoader(SmartField_FileLoader const&) = delete;
    void operator=(SmartField_FileLoader const&) = delete;

    QVector<QVector<uint8_t>> voltage;
    QVector<QVector<uint8_t>> current;
    //QList<QLineSeries*> voltage;
    //QList<QLineSeries*> current;
    void loadFile(QFile* inFile);

    float getHeaderVersion();
    float getNumberOfMeasurements();
    float getNumberOfPoints();
    QString getUID();
    float getNTCbeta();
    float getZsns();
    float getNormalisedFreqeuncy();
    void reset();

private:
    SmartField_FileLoader() {};
    float headerVersion = NAN;
    float numberOfMeasurements = NAN;
    float numberofPoints = NAN;
    QString UID;
    float NTCbeta = NAN;
    float Zsns = NAN;
    float normalisedFrequency = NAN;
    void processHeader(QByteArray* fileContent);
};

#endif // SMARTFIELD_FILELOADER_H
